// Code generated by "libovsdb.modelgen"
// DO NOT EDIT.

package ovsmodel

const NetFlowTable = "NetFlow"

// NetFlow defines an object in NetFlow table
type NetFlow struct {
	UUID             string            `ovsdb:"_uuid"`
	ActiveTimeout    int               `ovsdb:"active_timeout" validate:"min=-1"`
	AddIDToInterface bool              `ovsdb:"add_id_to_interface"`
	EngineID         *int              `ovsdb:"engine_id" validate:"omitempty,min=0,max=255"`
	EngineType       *int              `ovsdb:"engine_type" validate:"omitempty,min=0,max=255"`
	ExternalIDs      map[string]string `ovsdb:"external_ids"`
	Targets          []string          `ovsdb:"targets" validate:"min=1"`
}
