// Code generated by "libovsdb.modelgen"
// DO NOT EDIT.

package ovsmodel

const FlowTableTable = "Flow_Table"

type (
	FlowTableOverflowPolicy = string
)

var (
	FlowTableOverflowPolicyRefuse FlowTableOverflowPolicy = "refuse"
	FlowTableOverflowPolicyEvict  FlowTableOverflowPolicy = "evict"
)

// FlowTable defines an object in Flow_Table table
type FlowTable struct {
	UUID           string                   `ovsdb:"_uuid"`
	ExternalIDs    map[string]string        `ovsdb:"external_ids"`
	FlowLimit      *int                     `ovsdb:"flow_limit" validate:"omitempty,min=0"`
	Groups         []string                 `ovsdb:"groups"`
	Name           *string                  `ovsdb:"name"`
	OverflowPolicy *FlowTableOverflowPolicy `ovsdb:"overflow_policy" validate:"omitempty,oneof='refuse' 'evict'"`
	Prefixes       []string                 `ovsdb:"prefixes" validate:"max=3"`
}
